import random
from typing import List, Optional, Union

import streamlit as st
from bson.objectid import ObjectId
from pymongo import MongoClient

from xt_st_common.project_models import Project
from xt_st_common.session import get_user_email

from .config import StreamlitBaseSettings

settings = StreamlitBaseSettings()
PROJECT_COLLECTION_NAME = "project"


@st.cache_resource
def init_connection():
    settings = StreamlitBaseSettings()
    if None in [settings.DATABASE_NAME, settings.MONGO_CONNECTION_STRING]:
        raise ValueError("Can't get mongo engine when connection string is None.")
    client = MongoClient(settings.MONGO_CONNECTION_STRING)
    # engine = SyncEngine(client=client, database=settings.DATABASE_NAME)
    # engine.configure_database([Project], update_existing_indexes=True)
    return client.get_database(name=settings.DATABASE_NAME)


@st.cache_resource
def get_collection(name="project"):
    db_client = init_connection()
    return db_client[name]


@st.cache_data(ttl=30)
def get_project(project_id: str, cache_id=None) -> Optional[Project]:
    project_dict = get_collection().find_one({"_id": ObjectId(project_id)})
    return Project(**project_dict) if project_dict is not None else None


@st.cache_data(ttl=30)
def get_projects(include_public=False, cache_id=None) -> List[Project]:
    current_user = get_user_email()
    or_query = [
        {"users": {"$elemMatch": {"$regex": current_user, "$options": "i"}}},
        {"owner": {"$regex": current_user, "$options": "i"}},
    ]
    if include_public:
        or_query.append({"public": True})
    return [Project(**item) for item in get_collection().find({"$or": or_query})]


@st.cache_data(ttl=30)
def project_exists(project_id: str, cache_id=None) -> bool:
    return get_collection().count_documents({"_id": ObjectId(project_id)}) > 0


@st.cache_data(ttl=30)
def project_duplicate_exists(name: str, owner: str, project_id: Optional[str] = None, cache_id=None) -> bool:
    return (
        get_collection().count_documents(
            {"_id": {"$ne": ObjectId(project_id)}, "name": name, "owner": {"$regex": owner, "$options": "i"}}
        )
        > 0
    )


@st.cache_data(ttl=30)
def get_owned_projects(owner: str, cache_id=None) -> List[Project]:
    return [Project(**item) for item in get_collection().find({"owner": {"$regex": owner, "$options": "i"}})]


def reset_project_cache():
    st.session_state.project_cache = random.random()


def get_project_cache():
    if "project_cache" not in st.session_state:
        reset_project_cache()
    return st.session_state.project_cache


def save_project(project: Project):
    if project.id is None:
        result = get_collection().insert_one(project.dict(exclude={"id"}))
        project.id = result.inserted_id
    else:
        get_collection().update_one(
            {"_id": ObjectId(project.id) if project.id is not None else None},
            {"$set": project.dict(exclude={"id"})},
            upsert=True,
        )
    reset_project_cache()
    return project


def delete_project(project_id: Union[str, ObjectId]):
    get_collection().delete_one({"_id": ObjectId(project_id)})
    reset_project_cache()



""" An upload page which uses the server filesystem for storage """
import os
from datetime import datetime
from io import BytesIO
from pathlib import Path
from shutil import rmtree
from typing import Iterable, List, Union
from zipfile import ZIP_DEFLATED, ZipFile
from zoneinfo import ZoneInfo

import streamlit as st
from streamlit_tree_select import tree_select

from xt_st_common.config import StreamlitBaseSettings

settings = StreamlitBaseSettings()


def list_directory(path: str, include_empty=True) -> list[str]:
    """
    Return a list of all the files in a directory.
    """
    files = []
    for root, directories, filenames in os.walk(path):
        relative_root = root.replace(path, "")

        # Collect empty directories separately otherwise they never appear in the list
        if include_empty:
            for d in directories:
                full_dir = Path(root) / d
                if len(os.listdir(full_dir)) == 0:
                    files.append(str(Path(relative_root)) / d + os.sep)

        # Collect files
        for filename in filenames:
            files.append(Path(relative_root) / filename)
    return files


def create_file_tree(files: Iterable[str]) -> list[dict[str, str]]:
    """
    Using a list of file paths return a tree structure compatible with `streamlit_tree_select`.
    """
    file_tree = []
    files = sorted(files)
    for file in files:
        parts = str(file).split(os.sep)
        current_level = file_tree
        for i, part in enumerate(parts):
            existing = [node for node in current_level if node["label"] == part]
            if existing:
                current_level = existing[0].setdefault("children", [])
            else:
                new_node = {
                    "label": part,
                    "value": os.sep.join(parts[: i + 1]),
                }
                current_level.append(new_node)
                if i != len(parts) - 1:
                    current_level = new_node.setdefault("children", [])
    return file_tree


def _is_child_path(child_path, parent_path):
    """
    Returns `True` if the `child_path` is inside the `parent_path`.
    """
    # make both absolute
    parent_path = os.path.realpath(parent_path) + os.sep
    child_path = os.path.realpath(child_path)

    # return true, if the common prefix of both is equal to parent_path
    # e.g. /a/b/c/d.rst and directory is /a/b, the common prefix is /a/b
    return os.path.commonprefix([child_path, parent_path]) == parent_path


def _prepare_download(files: List[str], root_dir: str = "") -> BytesIO:
    """
    Prepare zip file for download
    """
    mem_zip = BytesIO()
    with ZipFile(mem_zip, "w") as zf:
        for file in files:
            fpath = Path(root_dir + os.sep + file)
            if fpath.is_dir():
                continue
            zf.write(str(fpath), arcname=file, compress_type=ZIP_DEFLATED)
    return mem_zip


def _delete_files(files: List[str], root_dir: str = "") -> None:
    """
    Delete the list of files. The root directory of a list of
    relative paths can be passed using `root_dir`.
    """
    dir_list = []
    for path in files:
        fpath = Path(root_dir + os.sep + path)

        # If you're deleting the root directory just recreate it rather than
        # manually recursing through the structure
        if path == "/" and len(root_dir) > 0:
            rmtree(root_dir)
            Path.mkdir(root_dir)
            return

        # Check path is a child of the expected root directory
        if not _is_child_path(fpath, root_dir):
            raise PermissionError(f"Attempting to remove file outside root directory: ['{path}']")

        # If path is a directory defer until later once all the files are removed
        # (can't remove a non-empty directory)
        if fpath.is_dir():
            dir_list.append(str(fpath))
        else:
            Path.unlink(fpath)

    # Remove the collected directories
    # - Sort longest->shortest so that we remove deepest directories first
    # - Create a set to make sure they're unique (empty directories appear twice otherwise)
    dir_list = sorted(set(dir_list), key=len, reverse=True)
    if "dir_list" not in st.session_state:
        st.session_state["dir_list"] = dir_list
    for path in dir_list:
        Path.rmdir(path)


def upload_page(data_dir: str, accepted_ft: Union[None, str, List[str]] = None, dl_prefix: str = "data"):
    """
    Render the upload page which uses the servers local filesystem for storage

    ### Parameters
        data_dir: str
            directory to store data in on the server

        accepted_ft: str or list of str or None
            Accepted filetypes for upload (passed to `st.file_uploader()`)

        dl_prefix: str
            Name of the zip when downloading will be `{dl_prefix}-Y-m-d_H.M.S.zip`
    """
    c1, c2 = st.columns(2)

    # Show file structure
    with c1:
        file_list = list_directory(data_dir + os.sep)
        file_tree = [{"label": "/", "value": "/", "children": create_file_tree(file_list)}]
        ts = tree_select(file_tree, expanded=["/"], check_model="all")

    # Show upload widget and other controls
    with c2:
        # File Upload
        uploaded_files = st.file_uploader("Upload input file(s)", accept_multiple_files=True, type=accepted_ft)
        for uploaded_file in uploaded_files:
            save_path = Path(data_dir, uploaded_file.name)
            with Path.open(save_path, "wb") as w:
                w.write(uploaded_file.getvalue())

        # File Download
        files_selected, dl_data = (False, "")
        if ts["checked"]:
            files_selected = True
            dl_data = _prepare_download(ts["checked"], root_dir=data_dir)

        l_btn, r_btn, _ = st.columns([1, 1, 2])
        with l_btn:
            st.download_button(
                "Zip & Download",
                data=dl_data,
                file_name=f"{dl_prefix}-{datetime.now(tz=ZoneInfo(settings.TIMEZONE)).strftime('%Y-%m-%d_%H.%M.%S')}.zip",
                disabled=not files_selected,
            )
        with r_btn:
            if st.button("Delete Selected", disabled=not files_selected, type="primary"):
                _delete_files(ts["checked"], root_dir=data_dir)
                st.experimental_rerun()



import logging
from enum import Enum
from pathlib import Path
from typing import Dict, List, Optional, Union

from bson import ObjectId
from pydantic import BaseModel, Field

from xt_st_common.config import StreamlitBaseSettings
from xt_st_common.storage import FileRef, storage_client

settings = StreamlitBaseSettings()


class PyObjectId(ObjectId):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if not ObjectId.is_valid(v):
            raise ValueError("Invalid objectid")
        return ObjectId(v)

    @classmethod
    def __modify_schema__(cls, field_schema):
        field_schema.update(type="string")


class ProjectState(str, Enum):
    PROJECT_DELETE_CONFIRM = "proj_delete_confirm"
    PROJECT_TO_DELETE = "proj_delete"
    PROJECT_SUCCESS_MESSAGE = "proj_success_message"
    PROJECT_WARNING_MESSAGE = "proj_warning_message"

    FOLDER_TO_DELETE = "upload_delete_folder"
    FOLDER_ADDED = "upload_folder_added"

    UPLOAD_SUCCESS_MESSAGE = "upload_success_message"
    UPLOAD_WARNING_MESSAGE = "upload_warning_message"
    UPLOAD_DELETE_CONFIRM = "upload_delete_confirm"

    FILE_DELETE_CONFIRM = "file_delete_confirm"
    FILE_TO_DELETE = "file_to_delete"
    FILE_SUCCESS_MESSAGE = "file_success_message"
    FILE_WARNING_MESSAGE = "file_warning_message"

    # FILE_MANAGER_UPLOAD = "file_manager_upload_file"
    FILE_MANAGER_REPLACE_FILE = "file_manager_replace_file"


class Project(BaseModel):
    id: Optional[PyObjectId] = Field(default_factory=PyObjectId, alias="_id")  # noqa: A003
    name: str
    description: str = ""
    public: bool = False
    application: str = settings.APP_NAME
    files: List[FileRef] = []
    folders: List[str] = []
    owner: str = ""
    users: List[str] = []  # Users with access to this project

    def get_users_string(self):
        return ",".join(self.users)

    def get_folders_map(self):
        fol_dict = {"/": "Project Root"}
        for folder in self.folders:
            parts = folder.strip("/").split("/")
            fol_dict[folder] = " - ".join(parts)
        return fol_dict

    def get_folder_path(self, folder):
        return f"{str(self.id)}/{self.name}/{folder}".strip("/")

    def get_files_in_folder(self, folder: str, include_subfolders=True, extensions=None, null_option=None):
        _files: Dict[str, Union[None, FileRef]] = {} if null_option is None else {null_option: None}

        path = self.get_folder_path(folder)
        for file in self.files:
            if path in file.path:
                file_path = file.path.removeprefix(path).strip("/")
                if (include_subfolders or file_path.find("/") <= 0) and (
                    extensions is None or str(Path(file.name).suffix.lower()) in extensions
                ):
                    _files[file_path] = file

        return _files

    def populate_node(self, _dict, part, parts):
        if part not in _dict:
            _dict[part] = {}

        if parts:
            self.populate_node(_dict[part], parts[0], parts[1:])

    def populate_children(self, _dict, key, children, path):
        if child_dict := _dict[key]:
            for child_key in child_dict:
                new_children = None
                new_path = f"{path}/{child_key}"
                if child_dict[child_key]:
                    new_children = []
                    self.populate_children(child_dict, child_key, new_children, new_path)
                children.append({"label": child_key, "value": f"{path}/{child_key}", "children": new_children})

    def get_file_tree(self):
        main_dict = {}
        full_list = (
            [file.path.removeprefix(self.get_folder_path("")).strip("/") for file in self.files] if self.files else []
        )
        if self.folders:
            full_list.extend(self.folders)
        if not full_list:
            return []
        for folder in full_list:
            parts = folder.split("/")
            self.populate_node(main_dict, parts[0], parts[1:])

        nodes = []
        for key in main_dict:
            children = None
            if main_dict[key]:
                children = []
                self.populate_children(main_dict, key, children, key)
            nodes.append({"label": key, "value": key, "children": children})

        return nodes

    def add_replace_file(self, data_file, folder: str, filename: str, content_type=None) -> FileRef:
        path = f"{self.get_folder_path(folder)}/{filename}".strip("/")
        return self.add_replace_file_by_path(data_file, path, content_type=content_type)

    def add_replace_file_by_path(self, data_file, path: str, content_type=None) -> FileRef:
        file_ref = storage_client().write_file(path, data_file, content_type)
        replaced = False
        for idx, file in enumerate(self.files):
            if file.path == path:
                self.files[idx] = file_ref
                replaced = True
                break
        if not replaced:
            self.files.append(file_ref)

        return file_ref

    def delete_file(self, file: FileRef):
        storage_client().delete_file(file.path)
        try:
            self.files.remove(file)
        except ValueError:
            logging.warning(f"File {file.path} didn't exist and couldn't be deleted.")

    def add_folders(self, folders_string: str):
        folders = set(self.folders)
        new_folders = folders_string.split(",")
        count = 0
        for folder in new_folders:
            folder = folder.strip().replace(" ", "_").strip("/")
            folder_parts = folder.split("/")
            for idx, part in enumerate(folder_parts):
                if idx == 0:
                    folders.add(f"{part}")
                else:
                    folders.add(f"{'/'.join(folder_parts[:idx+1])}")
                count += 1
        try:
            self.folders = list(folders)
        except AttributeError:
            logging.warning("Weird odmatic error, attempting to ignore")

        return count

    def delete_folder(self, folder: str):
        try:
            folder_path = self.get_folder_path(folder)
            self.folders.remove(folder)
            for file in [x for x in self.files if folder_path in x.path]:
                self.delete_file(file)
        except ValueError:
            logging.warning(f"Folder {folder} didn't exist and couldn't be deleted.")

    class Config:
        anystr_strip_whitespace = True
        # collection = "project"



import json
from io import BytesIO
from pathlib import Path
from random import randint
from typing import List, Literal, Optional, Tuple, Union  # , Optional

import pandas as pd
import streamlit as st

# from mpl_toolbox_ui.common.config import Settings
from xt_st_common.database import (
    Project,
    delete_project,
    get_project,
    get_project_cache,
    get_projects,
    project_duplicate_exists,
    save_project,
)
from xt_st_common.project_models import ProjectState
from xt_st_common.session import get_user_email
from xt_st_common.storage import FileRef, storage_client
from xt_st_common.utils import (
    get_encoding_and_dialect,
    get_state,
    seperate_users,
    set_state,
)

# from bson.objectid import ObjectId

# mapping for language options for streamlit code formatter. For other available options
# see: https://github.com/react-syntax-highlighter/react-syntax-highlighter/blob/master/AVAILABLE_LANGUAGES_PRISM.MD
CODE_FORMAT_MAPPING = {
    "json": "json",
    "yaml": "yaml",
    "yml": "yaml",
    "toml": "toml",
    "md": "markfown",
}

REPLACE_FILE_HELP_TXT = (
    "The new file can have a different name but must have the same extension."
    "Warning: Uploading a new file that is significantly different to the original "
    "can have catastrophic results."
)


def state_reset():
    for s in ProjectState:
        if s in [ProjectState.PROJECT_TO_DELETE, ProjectState.FILE_TO_DELETE]:
            set_state(s, None)
        else:
            set_state(s, "")


def get_selected_project() -> Optional[Project]:
    project = get_state("selected_project", None)
    if project is None:
        return None
    if not isinstance(project, Project):
        raise ValueError("Selected project is not the correct type")
    return project


def get_selected_project_or_error() -> Project:
    project = get_selected_project()
    if project is None:
        raise ValueError("No project was selected")
    return project


def set_selected_project(project: Optional[Project]):
    st.session_state.selected_project = project


def submit_delete_project(project: Project):
    """Callback function to set state in order to enable a delete on the next run."""
    # display a warning if the user entered an existing name

    if project is None:
        set_state(ProjectState.PROJECT_WARNING_MESSAGE, "No project is selected")
    else:
        set_state(
            ProjectState.PROJECT_DELETE_CONFIRM,
            (
                "Deleting will remove all files that are part of project: "
                + f"**'{project.name}'**. Are you sure you want to continue?"
            ),
        )
        set_state(ProjectState.PROJECT_TO_DELETE, project)


def action_delete(project: Project):
    if project.id is not None:
        delete_project(project.id)
    set_state(
        ProjectState.PROJECT_SUCCESS_MESSAGE,
        f"Project **'{project.name}'** deleted successfully",
    )


def submit_delete_folder(folder: str):
    """Callback function to set state in order to enable a delete on the next run."""
    # display a warning if the user entered an existing name
    project = st.session_state.selected_project
    if project is None:
        set_state(ProjectState.UPLOAD_WARNING_MESSAGE, "No project is selected")
    if folder is None:
        set_state(ProjectState.UPLOAD_WARNING_MESSAGE, "No folder is selected")
    else:
        set_state(
            ProjectState.UPLOAD_DELETE_CONFIRM,
            (
                "Deleting will remove all files that are part of this folder: "
                + f"**'{folder}'**. Are you sure you want to continue?"
            ),
        )
        set_state(ProjectState.FOLDER_TO_DELETE, folder)


def action_delete_folder(folder):
    project = get_selected_project_or_error()
    project.delete_folder(folder)
    save_project(project)

    set_state(ProjectState.FOLDER_ADDED, None)
    set_state(
        ProjectState.UPLOAD_SUCCESS_MESSAGE,
        f"Folders **'{folder}'** was deleted successfully",
    )


def submit_delete_file(file: FileRef):
    """Callback function to set state in order to enable a delete on the next run."""
    # display a warning if the user entered an existing name
    project = st.session_state.selected_project
    if project is None:
        set_state(ProjectState.FILE_WARNING_MESSAGE, "No project is selected")
    if file is None:
        set_state(ProjectState.FILE_WARNING_MESSAGE, "No file is selected")
    else:
        set_state(
            ProjectState.FILE_DELETE_CONFIRM,
            f"Are you sure you want to delete file **'{file.name}'**?",
        )
        set_state(ProjectState.FILE_TO_DELETE, file)


def action_delete_file(_file: FileRef):
    project = get_selected_project_or_error()
    project.delete_file(_file)
    save_project(project)
    set_state(ProjectState.FILE_SUCCESS_MESSAGE, f"File {_file.name} was deleted successfully")


def submit_add_project(project: Optional[Project] = None):
    """Callback function during adding a new project."""

    message_verb = "updated"
    container = st.session_state.message_box if "message_box" in st.session_state else st

    name = get_state("create_project_name")
    description = str(get_state("create_project_description"))
    users = get_state("create_project_users")
    is_public = bool(get_state("create_project_public", False))

    if not name:
        container.warning("No project name was provided")
        return

    users_list = seperate_users(users)
    if project is None:
        message_verb = "created"
        project = Project(
            name=name,
            owner=get_user_email(),
            description=description,
            users=users_list,
            public=is_public,
        )
        if project_duplicate_exists(project.name, project.owner, str(project.id)):
            st.session_state.proj_warning_message = f"A Project with the name: **'{name}'** already exists"
            return
    else:
        # Get latest DB copy of the project
        project = get_project(str(project.id), st.session_state.project_cache)
        if project is None:
            st.session_state.proj_warning_message = f"Cannot update project: **'{name}'** not found in database"
            return
        project.__dict__.update(
            {
                "name": name,
                "description": description,
                "users": users_list,
                "public": is_public,
            }
        )

    project = save_project(project)
    set_selected_project(project)

    set_state(
        ProjectState.PROJECT_SUCCESS_MESSAGE,
        f"Project **'{name}'** has been {message_verb}.",
    )
    return


def project_form(project: Optional[Project] = None):
    """A form that allows for creating an updating projects

    Parameters
    ----------
    project : Optional[Project], optional
        _description_, by default None
    """
    is_proj = project is not None
    name_val = project.name if is_proj else ""
    desc_val = project.description if is_proj else ""
    users_val = project.get_users_string() if is_proj else ""
    public_val = project.public if is_proj else False

    if is_proj:
        st.button(
            f"Delete {name_val}",
            on_click=submit_delete_project,
            args=(project,),
        )

    with st.form("create_project", clear_on_submit=True):
        if project is not None:
            st.subheader(f"Edit {project.name}")
        else:
            st.subheader("Create Project")
        st.text_input("Project Name", key="create_project_name", value=name_val)
        st.text_area("Description", key="create_project_description", value=desc_val)
        st.checkbox("Make Public", key="create_project_public", value=public_val)
        st.text_input("Users (emails , seperated)", key="create_project_users", value=users_val)

        st.form_submit_button(
            label="Submit",
            help="Create/Edit a Project",
            on_click=submit_add_project,
            args=(project,),
        )


def add_folders():
    project = st.session_state.selected_project
    folders_string = st.session_state.add_project_folder_name

    if not folders_string:
        set_state(
            ProjectState.UPLOAD_WARNING_MESSAGE,
            "Cannot add new folders: Folder name was empty",
        )
        return

    count = project.add_folders(folders_string)
    save_project(project)
    set_state(
        ProjectState.UPLOAD_SUCCESS_MESSAGE,
        f"{count} folders were added to **'{project.name}'**",
    )


@st.cache_data(ttl=300)
def get_df_preview(path: str, ext: Optional[str], num_rows=10):
    # if filepath.suffix == ".zip":
    #     frame = get_gdf_from_file(filepath)
    #     return frame.iloc[:num_rows, :-1]
    if ext == ".csv":
        file = storage_client().get_file(path)
        encoding, dialect = get_encoding_and_dialect(file)
        return pd.read_csv(file, nrows=num_rows, sep=dialect.delimiter, encoding=encoding)
    if ext == ".feather":
        file = storage_client().get_file(path)
        return pd.read_feather(file)

    return None


def get_string_preview(fileref: FileRef):
    file = storage_client().get_file(fileref.path)
    return file.getvalue().decode("utf-8")


def _state_name(project_id: str, folder: str) -> str:
    return f"{project_id}-{folder}_fs"


@st.cache_data(ttl=15)
def get_proj_options(include_public: bool, cache_id=None):
    selected_project = get_selected_project()
    projects = get_projects(include_public, get_project_cache())
    sel_idx = 0
    options = {}
    for idx, proj in enumerate(projects):
        if selected_project and proj.id == selected_project.id:
            sel_idx = idx + 1
        options[idx] = proj.name
    return options, projects, sel_idx


def on_project_select():
    proj_idx = get_state("project_select")
    include_public = bool(get_state("include_public_projects", False))
    options, projects, _ = get_proj_options(include_public, get_project_cache())

    if proj_idx is not None and proj_idx > -1:
        selected_project = projects[proj_idx] if proj_idx != -1 else None
        set_selected_project(selected_project)
    else:
        set_selected_project(None)


def project_selector(
    header_text: Optional[str] = "Projects",
    select_box_label="Select Project",
    null_option="-- Select Project --",
    st_context=st.sidebar,
) -> Tuple[Union["Project", None], List["Project"]]:
    """UI to select and create projects
    Args:
        root_path (Path): The root Path to where the project folders live
    """
    selected_project = None

    if header_text is not None:
        st_context.subheader(header_text)
    include_public = st_context.checkbox("Include Public", value=False, key="include_public_projects")
    options, projects, sel_idx = get_proj_options(include_public, get_project_cache())
    proj_options = {-1: null_option} if null_option is not None else {}
    if options is not None:
        proj_options = {**proj_options, **options}

    if len(proj_options) > 0:
        proj_idx = st_context.selectbox(
            select_box_label,
            key="project_select",
            # index=sel_idx,
            # on_change=on_project_select,
            options=proj_options.keys(),
            format_func=lambda x: proj_options[x],
        )

        selected_project = None
        if proj_idx is not None and proj_idx > -1:
            selected_project = projects[proj_idx] if proj_idx != -1 else None
            set_selected_project(selected_project)
    else:
        st_context.warning("No projects were found")

    return selected_project, projects


def load_csv(
    data_file,
    st_context=st,
):
    """
    Takes a csv file and loads it into session_state
    """

    try:
        encoding, dialect = get_encoding_and_dialect(data_file)
        raw_df = pd.read_csv(
            data_file,
            header=None,
            skip_blank_lines=True,
            engine="python",
            sep=None,
            encoding=encoding,
        )
    except Exception as err:
        raise ValueError("Could not parse txt/csv file.") from err

    c1, c2 = st.columns([1, 3])
    c1.header("CSV Data Import")
    c2.subheader("Import Preview (15 Rows)")
    if len(raw_df.columns) < 5:
        c2.info(
            "If preview has not loaded rows/columns correctly it may mean the wrong separator has been "
            + "detected. If that is the case than please check your file and remove unnecessary "
            + "header information."
        )
    c2.write(raw_df.head(15))

    with c1.form("config_df"):
        row_options: List[Union[str, int]] = list(range(16))
        row_options_wnone = row_options.copy()
        row_options_wnone.insert(0, "None")
        header_row = st_context.selectbox(label="Column Names Row", options=row_options)
        units_row = st_context.selectbox(label="Units Row", options=row_options_wnone)
        skip_rows = st_context.multiselect(label="Skip Rows", options=list(range(9)))

        if st.form_submit_button("Save Data"):
            return parse_csv_data(raw_df, header_row, units_row, skip_rows)
    return None, None


def parse_csv_data(raw_df: pd.DataFrame, header_row, units_row, skip_rows: Optional[List[int]] = None):
    units = {}
    if skip_rows is None:
        skip_rows = []

    raw_df.columns = raw_df.iloc[header_row]

    if units_row is not None and units_row != "None":
        units = raw_df.iloc[units_row].to_dict()
        if units_row != header_row and units_row not in skip_rows:
            raw_df = raw_df.drop(labels=units_row)

    if header_row not in skip_rows:
        raw_df = raw_df.drop(labels=header_row)

    raw_df = raw_df.drop(labels=skip_rows)
    raw_df = raw_df.reset_index(drop=True)

    return raw_df, units


def _update_key(prefix: str = "", replace: bool = False):
    """Hack to clear file upload after save by updating the key"""
    value = prefix + str(randint(1000, 100000000))
    set_state(f"{prefix}file_manager_key{'_replace' if replace else ''}", value)
    return str(value)


def _get_key(prefix: str = "", replace: bool = False):
    """Hack to clear file upload after save by updating the key"""
    key = get_state(f"{prefix}file_manager_key{'_replace' if replace else ''}", None)
    return _update_key(prefix, replace) if key is None else key


def file_manager(
    project: Project,
    types: List[str],
    label: str,
    st_context=st.sidebar,
    help_text: Optional[str] = None,
    allow_upload=True,
    allow_multiple_uploads=False,
    allow_delete=True,
    allow_replace=True,
    allow_folder_add=False,
    key_prefix: str = "",
    expand_file_actions=False,
    folder_select_text="Select Borehole/Run",
    auto_parse_csv=True,
    render_layout: Literal["vertical", "horizontal", "compact"] = "vertical",
):
    file_delete_confirm = get_state(ProjectState.FILE_DELETE_CONFIRM)
    file_to_delete = get_state(ProjectState.FILE_TO_DELETE)
    file_success_message = get_state(ProjectState.FILE_SUCCESS_MESSAGE)
    file_warning_message = get_state(ProjectState.FILE_WARNING_MESSAGE)

    folder_to_delete = get_state(ProjectState.FOLDER_TO_DELETE)
    folder_delete_confirm = get_state(ProjectState.UPLOAD_DELETE_CONFIRM)
    folder_success_message = get_state(ProjectState.UPLOAD_SUCCESS_MESSAGE)
    folder_warning_message = get_state(ProjectState.UPLOAD_WARNING_MESSAGE)

    st_context.subheader(f"Files: {project.name}")
    if file_delete_confirm and file_to_delete:
        st.warning(file_delete_confirm)
        st.button("I'm Sure", on_click=action_delete_file, args=(file_to_delete,))

    if file_success_message:
        st.success(file_success_message)
    if file_warning_message:
        st.warning(file_warning_message)

    if folder_delete_confirm and folder_to_delete:
        st.warning(folder_delete_confirm)
        st.button("I'm Sure", on_click=action_delete_folder, args=(folder_to_delete,))

    if folder_success_message:
        st.success(folder_success_message)
    if folder_warning_message:
        st.warning(folder_warning_message)

    if render_layout == "horizontal":
        container1, container2 = st_context.columns(2)
    else:
        container1 = st_context.container()
        container2 = st_context.container()

    folders_dict = project.get_folders_map()
    folder = container1.selectbox(
        folder_select_text,
        options=folders_dict.keys(),
        format_func=lambda x: folders_dict[x],
    )
    path = project.get_folder_path(folder) if folder else None
    if path is not None and folder is not None:
        row = container1.expander("Folder Actions", expanded=expand_file_actions)

        if allow_delete and folder != "/":
            row.button(
                "Delete Selected",
                key=f"{key_prefix}folder_delete_btn",
                on_click=submit_delete_folder,
                args=(folder,),
            )
        if allow_folder_add:
            row.caption("Add Sub Folders")
            row.text_input(
                "New Folders (use ',' to separate multiple folders and '/' to separate levels) ",
                key="add_project_folder_name",
                help="All folders are created relative to the project root. "
                + "Create multiple folders at once by using ',' as a separator, folders can be "
                + "multiple levels deep using '/' e.g. folder1/subfolder1, folder1/subfolder22",
                placeholder="folder1/subfolder1, folder1/subfolder2",
            )
            row.button(
                label="Add",
                help="Create new folder(s)",
                on_click=add_folders,
            )
        state = _state_name(str(project.id), folder)
        if state not in st.session_state:
            st.session_state[state] = 0

        if allow_upload:
            if auto_parse_csv:
                try_parse_csv = st.checkbox(
                    "Parse CSV/TXT as Dataset",
                    value=True,
                    help=(
                        "If a CSV or TXT file is uploaded you will be given options to help "
                        + "calibrate it for use as a dataset."
                    ),
                )
            else:
                try_parse_csv = False
            uploaded_files = container1.file_uploader(
                label,
                key=_get_key(key_prefix),
                type=types,
                help=help_text,
                accept_multiple_files=allow_multiple_uploads,
            )

            # handle cases where allow_multiple_uploads is false
            if uploaded_files is None:
                uploaded_files = []
            elif not isinstance(uploaded_files, list):
                uploaded_files = [uploaded_files]

            upload_messages = []
            for uploaded_file in uploaded_files:
                file_ref = None
                file_ref_units = None
                if (
                    uploaded_file
                    and try_parse_csv
                    and (uploaded_file.name.lower().endswith(".csv") or uploaded_file.name.lower().endswith(".txt"))
                ):
                    frame = None
                    try:
                        frame, units = load_csv(uploaded_file, st)
                    except ValueError:
                        st.warning(
                            "Could not parse CSV/TXT as a dataset. "
                            + "This may mean the file requires special parsing (such as a PWAVE file)"
                        )
                        try_parse_csv = not st.button("Upload anyway")
                        units = None

                    if frame is not None:
                        data_name = f"{Path(uploaded_file.name).stem}.feather"
                        with BytesIO() as buffer:
                            frame.to_feather(buffer)
                            buffer.seek(0)
                            file_ref = project.add_replace_file(
                                buffer,
                                folder=folder,
                                filename=data_name,
                            )
                        units_name = ""
                        if units:
                            units_name = f"{Path(uploaded_file.name).stem}_units.json"
                            units_string = json.dumps(units)
                            file_ref_units = project.add_replace_file(
                                units_string,
                                folder=folder,
                                filename=units_name,
                                content_type="application/json",
                            )

                elif uploaded_file:
                    file_ref = project.add_replace_file(
                        uploaded_file.getvalue(),
                        folder=folder,
                        filename=uploaded_file.name,
                    )

                if uploaded_file and file_ref:
                    uploaded_file.close()
                    upload_messages.append(
                        f"File: **'{file_ref.name}'** {' and ' + file_ref_units.name if file_ref_units else ''} "
                        + " uploaded successfully",
                    )

            if len(upload_messages) > 0:
                save_project(project)
                _update_key(key_prefix)
                set_state(
                    ProjectState.FILE_SUCCESS_MESSAGE,
                    "".join([f"> 1. {msg} \n" for msg in upload_messages]),
                )
                st.experimental_rerun()

        files = project.get_files_in_folder(folder)
        if files is not None and len(files) > 0:
            selected_key = container2.selectbox(
                "Select File",
                options=files.keys(),
                key=f"{key_prefix}file_manager_file_select",
            )
            selected_file = files[selected_key] if selected_key in files else None
            if selected_file is not None and selected_key is not None:
                row = container2.expander("File Actions", expanded=expand_file_actions)
                if len(selected_key) > 30:
                    row.caption(selected_key)
                # options = []
                if allow_delete:
                    row.button(
                        "Delete",
                        key=f"{key_prefix}file_delete_btn",
                        on_click=submit_delete_file,
                        args=(selected_file,),
                    )
                if selected_key.lower().endswith((".zip", ".csv", ".geojson", ".gpkg", ".feather")):
                    preview_frame = row.button(
                        "Preview Frame",
                        key=f"{key_prefix}file_manager_preview_frame",
                    )
                    if preview_frame:
                        with st.expander(f"**Frame Viewer:** {selected_file.name}", expanded=True):
                            st.dataframe(get_df_preview(selected_file.path, selected_file.get_ext()))
                if selected_key.lower().endswith((".json", ".yml", ".yaml", ".toml", ".md", ".txt")) and (
                    preview_frame := row.button(
                        "Preview File",
                        key=f"{key_prefix}file_manager_preview_file",
                    )
                ):
                    with st.expander(f"**File Viewer:** {selected_file.name}", expanded=True):
                        code_format = CODE_FORMAT_MAPPING.get(selected_key.lower().split(".")[-1])
                        if code_format is None:
                            st.write(get_string_preview(selected_file))
                        else:
                            st.code(
                                get_string_preview(selected_file),
                                language=code_format,
                            )
                if row.checkbox(
                    "Prepare Download",
                    key=f"{key_prefix}file_manager_download_chbx",
                ):
                    file_data = storage_client().get_file(selected_file.path)

                    row.download_button(
                        "Download",
                        file_data,
                        selected_file.name,
                        key=f"{key_prefix}file_manager_download_button",
                    )
                if allow_replace and (
                    uploaded_replace_file := row.file_uploader(
                        "Replace the selected file",
                        key=_get_key(key_prefix, True),
                        type=selected_file.get_ext(),
                        help=REPLACE_FILE_HELP_TXT,
                        accept_multiple_files=False,
                    )
                ):
                    project.add_replace_file_by_path(uploaded_replace_file, selected_file.path)
                    save_project(project)
                    uploaded_replace_file.close()
                    _update_key(key_prefix, True)
                    st.experimental_rerun()
        else:
            container2.markdown("##")
            container2.info("No files in selected folder")

        state_reset()
    return path, folder



from enum import Enum
from importlib.machinery import PathFinder
from pathlib import Path
from typing import Optional

from pydantic import BaseSettings, SecretStr
from pydantic.color import Color


class StorageType(str, Enum):
    MINIO = "MINIO"
    MINIO_AWS = "MINIO_AWS"
    NONE = "NONE"


class StreamlitBaseSettings(BaseSettings):
    SIGNOUT_URL: Optional[str] = None
    CURRENT_PACKAGE = __name__.split(".")[0]
    BASE_PATH: str = str(Path(PathFinder().find_spec(CURRENT_PACKAGE).origin).parent)  # type: ignore
    DATA_DIR: str = str(Path.cwd() / "appdata")
    """Root directory for storing app data"""

    ###############
    # Storage Vars
    ###############
    STORAGE_TYPE: StorageType = StorageType.NONE
    BUCKET_NAME: Optional[str] = None
    BUCKET_PREFIX: Optional[str] = None
    MINIO_SECRET_KEY: Optional[SecretStr] = None
    MINIO_ACCESS_KEY: Optional[str] = None
    MINIO_ENDPOINT: Optional[str] = None
    MINIO_HTTPS: bool = True

    APP_NAME: str = "my_app"
    """Name of the app that be tagged in the DB"""
    APP_TAG_TEXT: str = "Development"
    """text that will appear next the application logo"""
    APP_TAG_BACKGROUND: Color = Color("#00a9ce")
    """background colour of the APP_TAG_TEXT"""
    DEBUG: bool = False
    """Enable debug information in the app"""
    DEBUG_MOCK_SESSION: bool = False
    """Mock the session headers returned when deployed to AWS"""
    USE_COGNITO: bool = False
    """Use AWS Cognito for Auth"""
    COGNITO_COOKIE: str = "AWSELBAuthSessionCookie"
    """If set cookies with this prefix will be deleted on signout"""
    COGNITO_GROUPS: str = "ap-southeast-2_lC6GUKOej_CSIRO-EASI"
    """One of these comma separated groups must be present in the user's `cognito:groups`"""
    NO_ACCESS_MSG: str = "Error: You do not have permission to use this app."
    """Error message displayed when a user doesn't have permission to use the app"""
    STREAMLIT_SERVER_BASE_URL_PATH: str = ""
    """The base path for the URL where Streamlit should be served from"""
    MONGO_CONNECTION_STRING: str = ""
    """Mongo DB connection string"""
    DATABASE_NAME: str = ""
    """Mongo DB Database Name"""
    TIMEZONE: str = "Australia/Perth"
    """Default timezone used within the app"""

    class Config:
        env_file = ".env"
        env_file_encoding = "utf-8"
        case_sensitive = True



import streamlit as st

from xt_st_common.config import StreamlitBaseSettings

settings = StreamlitBaseSettings()


def page_header(
    page_title: str | None = None,
    page_logo: str | None = None,
    page_logo_width: int = 250,
    logout_url: str | None = None,
):
    """
    Display the page header including Title, Logo, and optional Logout button
    """
    base_url = settings.STREAMLIT_SERVER_BASE_URL_PATH
    script_frame = st.empty()
    if logout_url is not None:
        # Load the css/icons for the button
        st.markdown(
            """
            <style scoped type="text/css">
                .btn-logout {
                    box-shadow:inset 0px 1px 0px 0px #ffffff;
                    background:linear-gradient(to bottom, #f9f9f9 5%, #e9e9e9 100%);
                    background-color:#f9f9f9;
                    border-radius:6px;
                    border:1px solid #dcdcdc;
                    display:inline-block;
                    cursor:pointer;
                    color:#666666 !important;
                    font-family:Arial;
                    font-size:15px;
                    padding:4px 20px !important;
                    text-decoration:none;
                    text-shadow:0px 1px 0px #ffffff;
                }
                .btn-logout:hover {
                    background:linear-gradient(to bottom, #e9e9e9 5%, #f9f9f9 100%);
                    background-color:#e9e9e9;
                }
                .btn-logout:active {
                    position:relative;
                    top:1px;
                }
                .main .block-container {
                    padding-top: 3rem;
                }
            </style>
            """,
            unsafe_allow_html=True,
        )

        # Define the button
        logout_link = f'<a id="logout-button" class="btn-logout" href={logout_url} role="button">🔐 Log out</a>'
    else:
        logout_link = "&nbsp;"

    page_title = "" if page_title is None else page_title

    page_logo = "" if page_logo is None else f"<img src='{page_logo}' width='{page_logo_width}px'>"
    # Display the page header including Title, Logo, and optional Logout button
    st.markdown(
        f"""
        <style scoped type="text/css">
            .xt_logo {{
                border-bottom: 1px solid black;
                width: 100%;
                height: 65px;
                margin-bottom: 12px
            }}
            .app_logo{{
                font-family: Consolas, monaco, monospace;
                font-size:200%;
                font-weight:500;
            }}
            .app_logo::after {{
                font-family: sans-serif;
                font-size: small;
                content: '{settings.APP_TAG_TEXT}';
                display: inline-block;
                text-align: center;
                color: white;
                background-color: {settings.APP_TAG_BACKGROUND};
                border-radius: 20px;
                margin-left: -15px;
                padding: 0px 15px;
                font-weight: bold;
                vertical-align: super;
            }}
        </style>
        """
        + f"""
        <div class='xt_logo' >
            <span style='float: right;'>
            {logout_link}
            <img src='{base_url}/app/static/media/XT_words_wrapped.png' width='200px'>
            </span>
            <span class='app_logo'>
        """
        + page_logo
        + page_title
        + """
         </span>
        </div>
        """,
        unsafe_allow_html=True,
    )

    # Attach a function to clear the Cognito cookies when the logout button is clicked
    if settings.USE_COGNITO and len(settings.COGNITO_COOKIE) > 0:
        from streamlit_js_eval import streamlit_js_eval

        # Attach and onClick event handler to the logout button/link
        # When the user logs out the handler will remove cookies which are prefixed with
        # `settings.COGNITO_COOKIE`.
        # Also take the opportunity to have the signout button use the current window rather
        # than opening a new one.
        logout_script = (
            """
                var a = window.parent.document.getElementById("logout-button");
                a.onclick = function() {
                    all_cookies = Object.fromEntries(document.cookie.split('; ').map(c => c.split('=')))
                    cookie_list = Object.keys(all_cookies).filter(key => key.startsWith('"""
            + settings.COGNITO_COOKIE
            + """'))
                    cookie_list.forEach(e => {
                        window.parent.document.cookie = e+'=; expires=Thu, 01 Jan 1970 00:00:01 GMT'
                    })
                    return true;
                }
                // Override the target so that we reload the current page rather than opening a new one
                a.target = ""
            """
        )
        with script_frame:
            streamlit_js_eval(js_expressions=logout_script, key="js-logout-script")



import os
from json import dumps, loads
from pathlib import Path

import streamlit.components.v1 as components

# Create a _RELEASE constant. We'll set this to False while we're developing
# the component, and True when we're ready to package and distribute it.
# (This is, of course, optional - there are innumerable ways to manage your
# release process.)
_RELEASE = True

# Declare a Streamlit component. `declare_component` returns a function
# that is used to create instances of the component. We're naming this
# function "_component_func", with an underscore prefix, because we don't want
# to expose it directly to users. Instead, we will create a custom wrapper
# function, below, that will serve as our component's public API.

# It's worth noting that this call to `declare_component` is the
# *only thing* you need to do to create the binding between Streamlit and
# your component frontend. Everything else we do in this file is simply a
# best practice.

if not _RELEASE:
    _component_func = components.declare_component(
        # We give the component a simple, descriptive name ("my_component"
        # does not fit this bill, so please choose something better for your
        # own component :)
        "plotly_events",
        # Pass `url` here to tell Streamlit that the component will be served
        # by the local dev server that you run via `npm run start`.
        # (This is useful while your component is in development.)
        url="http://localhost:3001",
    )
else:
    # When we're distributing a production version of the component, we'll
    # replace the `url` param with `path`, and point it to to the component's
    # build directory:
    parent_dir = Path(__file__).resolve().parent
    build_dir = parent_dir / "frontend/build"
    _component_func = components.declare_component("plotly_events", path=str(build_dir))


# Create a wrapper function for the component. This is an optional
# best practice - we could simply expose the component function returned by
# `declare_component` and call it done. The wrapper allows us to customize
# our component's API: we can pre-process its input args, post-process its
# output value, and add a docstring for users.
def plotly_events(
    plot_fig,
    click_event=True,
    select_event=False,
    hover_event=False,
    override_height=450,
    override_width="100%",
    config={},  # noqa: B006
    key=None,
):
    """Create a new instance of "plotly_events".

    Parameters
    ----------
    plot_fig: Plotly Figure
        Plotly figure that we want to render in Streamlit
    click_event: boolean, default: True
        Watch for click events on plot and return point data when triggered
    select_event: boolean, default: False
        Watch for select events on plot and return point data when triggered
    hover_event: boolean, default: False
        Watch for hover events on plot and return point data when triggered
    override_height: int, default: 450
        Integer to override component height.  Defaults to 450 (px)
    override_width: string, default: '100%'
        String (or integer) to override width.  Defaults to 100% (whole width of iframe)
    config: dictionary, default: {}
        Plotly config dictionary. Defaults to {}
    key: str or None
        An optional key that uniquely identifies this component. If this is
        None, and the component's arguments are changed, the component will
        be re-mounted in the Streamlit frontend and lose its current state.

    Returns
    -------
    list of dict
        List of dictionaries containing point details (in case multiple overlapping
        points have been clicked).

        Details can be found here:
            https://plotly.com/javascript/plotlyjs-events/#event-data

        Format of dict:
            {
                x: int (x value of point),
                y: int (y value of point),
                curveNumber: (index of curve),
                pointNumber: (index of selected point),
                pointIndex: (index of selected point)
            }

    """
    # kwargs will be exposed to frontend in "args"
    component_value = _component_func(
        plot_obj=plot_fig.to_json(),
        override_height=override_height,
        override_width=override_width,
        config=dumps(config),
        key=key,
        click_event=click_event,
        select_event=select_event,
        hover_event=hover_event,
        default="[]",  # Default return empty JSON list
    )

    # Parse component_value since it's JSON and return to Streamlit
    return loads(component_value)


# Add some test code to play with the component while it's in development.
# During development, we can run this just as we would any other Streamlit
# app: `$ streamlit run src/streamlit_plotly_events/__init__.py`
if not _RELEASE:
    import plotly.express as px
    import streamlit as st

    st.set_page_config(layout="wide")

    st.subheader("Plotly Line Chart")
    fig = px.line(x=[0, 1, 2, 3], y=[0, 1, 2, 3])
    plot_name_holder = st.empty()
    clickedPoint = plotly_events(fig, key="line", config={"displayModeBar": True})
    plot_name_holder.write(f"Clicked Point: {clickedPoint}")

    # Here we add columns to check auto-resize/etc
    st.subheader("Plotly Bar Chart (With columns)")
    _, c2, _ = st.beta_columns((1, 6, 1))
    with c2:
        fig2 = px.bar(x=[0, 1, 2, 3], y=[0, 1, 2, 3])
        plot_name_holder2 = st.empty()
        clickedPoint2 = plotly_events(fig2, key="bar")
        plot_name_holder2.write(f"Clicked Point: {clickedPoint2}")

    st.subheader("# Plotly Select Event")
    fig3 = px.bar(x=[0, 1, 2, 3], y=[0, 1, 2, 3])
    plot_name_holder3 = st.empty()
    clickedPoint3 = plotly_events(fig3, key="select", click_event=False, select_event=True)
    plot_name_holder3.write(f"Selected Point: {clickedPoint3}")

    st.subheader("# Plotly Hover Event")
    fig4 = px.bar(x=[0, 1, 2, 3], y=[0, 1, 2, 3])
    plot_name_holder4 = st.empty()
    clickedPoint4 = plotly_events(fig4, key="hover", click_event=False, hover_event=True)
    plot_name_holder4.write(f"Hovered Point: {clickedPoint4}")



import csv
from collections import defaultdict
from enum import Enum
from io import BytesIO
from typing import Any, Optional, Union

import chardet
import streamlit as st

# from pydantic import BaseModel

# from dask.distributed import Client

# @st.cache(allow_output_mutation=True, hash_funcs={"_thread.RLock": lambda _: None})
# def get_dask_client():
#     client = Client("tcp://127.0.0.1:8786")
#     # client = Client()

#     return client


def get_encoding_and_dialect(input_file: BytesIO):
    sample_bytes = 1048576  # 1 Megabyte
    _bytes = input_file.read(sample_bytes)

    result = chardet.detect(_bytes)
    charenc = result["encoding"]

    if charenc == "ascii":
        charenc = "utf8"

    dialect = csv.Sniffer().sniff(str(_bytes))
    input_file.seek(0)

    return charenc, dialect


def sizeof_fmt(num: Optional[Union[float, int]], suffix="B") -> str:
    if num is None:
        return ""
    for unit in ["", "K", "M", "G", "T", "P", "E", "Z"]:
        if abs(num) < 1024.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f%s%s" % (num, "Y", suffix)


def seperate_users(users_string):
    users_array = users_string.strip().lower().split(",")
    return [user.strip() for user in users_array]


FILE_MARKER = "<files>"


def attach(branch, trunk):
    """
    Insert a branch of directories on its trunk.
    """
    parts = branch.split("/", 1)
    if len(parts) == 1:  # branch is a file
        trunk[FILE_MARKER].append(parts[0])
    else:
        node, others = parts
        if node not in trunk:
            trunk[node] = defaultdict(dict, ((FILE_MARKER, []),))
        attach(others, trunk[node])


def get_state(state_name: Union[str, Enum], default: Optional[Any] = ""):
    return st.session_state[str(state_name)] if str(state_name) in st.session_state else default


def set_state(state_name: Union[str, Enum], value: Any):
    st.session_state[str(state_name)] = value



from pathlib import Path
from typing import List, Optional

import jwt
from streamlit.web.server.websocket_headers import _get_websocket_headers

from xt_st_common.config import StreamlitBaseSettings

settings = StreamlitBaseSettings()


def get_session_headers():
    """
    Returns the request headers for the current streamlit session
    """
    if settings.DEBUG and settings.DEBUG_MOCK_SESSION:
        headers = {
            "X-Forwarded-For": "130.116.147.135",
            "X-Forwarded-Proto": "https",
            "X-Forwarded-Port": "443",
            "Host": "fracg-ui.exploration.tools",
            "X-Amzn-Trace-Id": "Root=1-642250e6-3b44a7d41992a61831721a43",
            "X-Amzn-Oidc-Data": "eyJ0eXAiOiJKV1QiLCJraWQiOiI2NGYyMTU4MS01MzEzLTRiZTUtYTVlYi1hMjUxMDliYTE0NTIiLCJhbGciOiJFUzI1NiIsImlzcyI6Imh0dHBzOi8vY29nbml0by1pZHAuYXAtc291dGhlYXN0LTIuYW1hem9uYXdzLmNvbS9hcC1zb3V0aGVhc3QtMl9sQzZHVUtPZWoiLCJjbGllbnQiOiIyaDhzM2NsZWMxNWgyaDRydTZnaDQ3MzdxMyIsInNpZ25lciI6ImFybjphd3M6ZWxhc3RpY2xvYWRiYWxhbmNpbmc6YXAtc291dGhlYXN0LTI6NDQ0NDg4MzU3NTQzOmxvYWRiYWxhbmNlci9hcHAvazhzLXh0YXBwc3Byb2QtMmYxYjk5ZGMwZC9jMjJlNTYxOTY0NjU2Y2ExIiwiZXhwIjoxNjc5OTcwNjU0fQ==.eyJzdWIiOiJiNWFiY2QyMy00NGI0LTQ4OTctOGFlOS00N2U1YjM0ODRjYTUiLCJuYW1lIjoiQWxleCBIdW50IiwiZW1haWwiOiJhbGV4Lmh1bnRAY3Npcm8uYXUiLCJ1c2VybmFtZSI6Imh1bjIyMCIsImV4cCI6MTY3OTk3MDY1NCwiaXNzIjoiaHR0cHM6Ly9jb2duaXRvLWlkcC5hcC1zb3V0aGVhc3QtMi5hbWF6b25hd3MuY29tL2FwLXNvdXRoZWFzdC0yX2xDNkdVS09laiJ9.z7LBzOiuudjAToQJIxQNrOHvr_CgwtcHMJx0nsg7qcTKuFcna2GGcZGcMBal9KTnF6YEb-bVh8XmH3Y4lGR9SA==",  # noqa: E501
            "X-Amzn-Oidc-Identity": "b5abcd23-44b4-4897-8ae9-47e5b3484ca5",
            "X-Amzn-Oidc-Accesstoken": "eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.eyJpc3MiOiJodHRwczovL2NvZ25pdG8taWRwLmFwLXNvdXRoZWFzdC0yLmFtYXpvbmF3cy5jb20vRVhBTVBMRSIsImlhdCI6MTY3OTk4NzU3MSwiZXhwIjo0MDc4Mjc4NzcxLCJhdWQiOiJ3d3cuZXhhbXBsZS5jb20iLCJzdWIiOiIxMTExY2QyMy00NGI0LTQ4OTctOGFlOS00N2U1YjM0ODExMTEiLCJjb2duaXRvOmdyb3VwcyI6WyJhcHAtZnJhY2ciLCJQcm9qZWN0IEFkbWluaXN0cmF0b3IiXSwidmVyc2lvbiI6IjIiLCJjbGllbnRfaWQiOiIxMTExMTExMTExMTExMTExMTExMTExMTExMSIsInRva2VuX3VzZSI6ImFjY2VzcyIsInNjb3BlIjoiYXdzLmNvZ25pdG8uc2lnbmluLnVzZXIuYWRtaW4gb3BlbmlkIHByb2ZpbGUgZW1haWwiLCJ1c2VybmFtZSI6Inp6ejk5OSIsImp0aSI6IjExMTE0YTJhLWQ5NmEtMTExMS05MWNhLTlmMTBiNjE2MjVjZSIsImF1dGhfdGltZSI6IjE2Nzk5ODc1NzEiLCJvcmlnaW5fanRpIjoiM2E2MDQxNDAtMGQ5NC0xMTExLTExMTEtMDY0MzQ4MmMxMTExIn0.al-W2v2WgbeBpOvEY-p8KQAqjlUn1TMPvw5rgQtYe4o",  # noqa: E501
            "Upgrade": "websocket",
            "Connection": "upgrade",
            "Pragma": "no-cache",
            "Cache-Control": "no-cache",
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/111.0.0.0 Safari/537.36",  # noqa: E501
            "Origin": "https://fracg-ui.exploration.tools",
            "Sec-Websocket-Version": "13",
            "Accept-Encoding": "gzip, deflate, br",
            "Accept-Language": "en-GB,en-US;q=0.9,en;q=0.8",
            "Cookie": "",
            "Sec-Websocket-Key": "t7f+56WSS5xn+qdI4L4Yqg==",
            "Sec-Websocket-Extensions": "permessage-deflate; client_max_window_bits",
            "Sec-Websocket-Protocol": "streamlit",
        }
        return headers

    return _get_websocket_headers()


def decode_jwt(raw_jwt: str, algorithms: Optional[List[str]] = None, options=None) -> dict:
    """
    Decode a JWT and return its values as a dictionary

    `algorithms` defaults to `["ES256"]`
    `options` defaults to `{"verify_signature": False}`
    """
    if algorithms is None:
        algorithms = ["ES256"]
    if options is None:
        options = {"verify_signature": False}
    return jwt.decode(raw_jwt, algorithms=algorithms, options=options)


def get_user_id(allow_default=True):
    """
    Returns the current user id

    allow_default:
        return "default-user" if user not found
    """

    headers = get_session_headers()
    if headers is not None:
        if "X-Auth-Request-User" in headers:
            return headers["X-Auth-Request-User"]
        if "X-Amzn-Oidc-Data" in headers:
            return decode_jwt(headers["X-Amzn-Oidc-Data"])["sub"]
        if "X-Goog-Authenticated-User-Id" in headers:
            return headers["X-Goog-Authenticated-User-Id"].split(":")[1]
    if allow_default:
        return "default-user"

    raise KeyError("X-Auth-Request-User not found in headers")


def get_user_email(allow_default=True):
    """
    Return the user's email address from headers, otherwise return the string "default-user"
    """
    headers = get_session_headers()
    if headers is not None and "X-Goog-Authenticated-User-Email" in headers:
        return headers["X-Goog-Authenticated-User-Email"].split(":")[1]
    if headers is not None and "X-Amzn-Oidc-Data" in headers:
        return decode_jwt(headers["X-Amzn-Oidc-Data"])["email"]
    if allow_default:
        return "default-user@exploration.tools"

    raise KeyError("X-Goog-Authenticated-User-Email not found in headers")


def get_oidc_groups() -> List[str]:
    """
    Attempts to retrieve the OIDC groups from the headers (e.g. cognito:groups)
    """
    headers = get_session_headers()
    if headers is None or "X-Amzn-Oidc-Accesstoken" not in headers:
        raise KeyError("X-Amzn-Oidc-Accesstoken not found in headers")
    access_tok = decode_jwt(headers["X-Amzn-Oidc-Accesstoken"])
    return access_tok.get("cognito:groups", [])


def get_user_name(allow_default=True):
    """
    Return user's name from headers, otherwise return the string "default-user"
    """
    headers = get_session_headers()
    if headers is not None and "X-Amzn-Oidc-Data" in headers:
        return decode_jwt(headers["X-Amzn-Oidc-Data"])["name"]
    if allow_default:
        return "default-user"

    raise KeyError("X-Amzn-Oidc-Data not found in headers")


def get_user_path() -> str:
    """
    Return path to local file storage for the user
    """
    return str(Path(settings.DATA_DIR) / get_user_email())
